@@ -1,5 +1,5 @@
-# dialects/mysql/reflection.py
-# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
+# mysql/reflection.py
+# Copyright (C) 2005-2022 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -322,12 +322,7 @@ class MySQLTableDefinitionParser(object):
     def _parse_keyexprs(self, identifiers):
         """Unpack '"col"(2),"col" ASC'-ish strings into components."""
 
-        return [
-            (colname, int(length) if length else None, modifiers)
-            for colname, length, modifiers in self._re_keyexprs.findall(
-                identifiers
-            )
-        ]
+        return self._re_keyexprs.findall(identifiers)
 
     def _prep_regexes(self):
         """Pre-compile regular expressions."""
