@@ -1,5 +1,5 @@
 # engine/row.py
-# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2022 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -66,21 +66,25 @@ except ImportError:
         def __init__(self, parent, processors, keymap, key_style, data):
             """Row objects are constructed by CursorResult objects."""
 
-            self._parent = parent
+            object.__setattr__(self, "_parent", parent)
 
             if processors:
-                self._data = tuple(
-                    [
-                        proc(value) if proc else value
-                        for proc, value in zip(processors, data)
-                    ]
+                object.__setattr__(
+                    self,
+                    "_data",
+                    tuple(
+                        [
+                            proc(value) if proc else value
+                            for proc, value in zip(processors, data)
+                        ]
+                    ),
                 )
             else:
-                self._data = tuple(data)
+                object.__setattr__(self, "_data", tuple(data))
 
-            self._keymap = keymap
+            object.__setattr__(self, "_keymap", keymap)
 
-            self._key_style = key_style
+            object.__setattr__(self, "_key_style", key_style)
 
         def __reduce__(self):
             return (
@@ -187,7 +191,7 @@ class Row(BaseRow, collections_abc.Sequence):
 
     .. seealso::
 
-        :ref:`coretutorial_selecting` - includes examples of selecting
+        :ref:`tutorial_selecting_data` - includes examples of selecting
         rows from SELECT statements.
 
         :class:`.LegacyRow` - Compatibility interface introduced in SQLAlchemy
@@ -211,6 +215,12 @@ class Row(BaseRow, collections_abc.Sequence):
     # in 2.0, this should be KEY_INTEGER_ONLY
     _default_key_style = KEY_OBJECTS_BUT_WARN
 
+    def __setattr__(self, name, value):
+        raise AttributeError("can't set attribute")
+
+    def __delattr__(self, name):
+        raise AttributeError("can't delete attribute")
+
     @property
     def _mapping(self):
         """Return a :class:`.RowMapping` for this :class:`.Row`.
@@ -269,10 +279,11 @@ class Row(BaseRow, collections_abc.Sequence):
         }
 
     def __setstate__(self, state):
-        self._parent = parent = state["_parent"]
-        self._data = state["_data"]
-        self._keymap = parent._keymap
-        self._key_style = state["_key_style"]
+        parent = state["_parent"]
+        object.__setattr__(self, "_parent", parent)
+        object.__setattr__(self, "_data", state["_data"])
+        object.__setattr__(self, "_keymap", parent._keymap)
+        object.__setattr__(self, "_key_style", state["_key_style"])
 
     def _op(self, other, op):
         return (
@@ -538,7 +549,8 @@ class ROMappingView(
 
 
 class RowMapping(BaseRow, collections_abc.Mapping):
-    """A ``Mapping`` that maps column names and objects to :class:`.Row` values.
+    """A ``Mapping`` that maps column names and objects to :class:`.Row`
+    values.
 
     The :class:`.RowMapping` is available from a :class:`.Row` via the
     :attr:`.Row._mapping` attribute, as well as from the iterable interface
