@@ -1,9 +1,9 @@
 # ext/associationproxy.py
-# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2022 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
-# the MIT License: http://www.opensource.org/licenses/mit-license.php
+# the MIT License: https://www.opensource.org/licenses/mit-license.php
 
 """Contain the ``AssociationProxy`` class.
 
@@ -383,6 +383,22 @@ class AssociationProxyInstance(object):
             return AmbiguousAssociationProxyInstance(
                 parent, owning_class, target_class, value_attr
             )
+        except Exception as err:
+            util.raise_(
+                exc.InvalidRequestError(
+                    "Association proxy received an unexpected error when "
+                    "trying to retreive attribute "
+                    '"%s.%s" from '
+                    'class "%s": %s'
+                    % (
+                        target_class.__name__,
+                        parent.value_attr,
+                        target_class.__name__,
+                        err,
+                    )
+                ),
+                from_=err,
+            )
         else:
             return cls._construct_for_assoc(
                 target_assoc, parent, owning_class, target_class, value_attr
@@ -472,10 +488,23 @@ class AssociationProxyInstance(object):
     def attr(self):
         """Return a tuple of ``(local_attr, remote_attr)``.
 
-        This attribute is convenient when specifying a join
-        using :meth:`_query.Query.join` across two relationships::
+        This attribute was originally intended to facilitate using the
+        :meth:`_query.Query.join` method to join across the two relationships
+        at once, however this makes use of a deprecated calling style.
 
-            sess.query(Parent).join(*Parent.proxied.attr)
+        To use :meth:`_sql.select.join` or :meth:`_orm.Query.join` with
+        an association proxy, the current method is to make use of the
+        :attr:`.AssociationProxyInstance.local_attr` and
+        :attr:`.AssociationProxyInstance.remote_attr` attributes separately::
+
+            stmt = (
+                select(Parent).
+                join(Parent.proxied.local_attr).
+                join(Parent.proxied.remote_attr)
+            )
+
+        A future release may seek to provide a more succinct join pattern
+        for association proxy attributes.
 
         .. seealso::
 
@@ -1192,7 +1221,7 @@ class _AssociationList(_AssociationCollection):
 
     for func_name, func in list(locals().items()):
         if (
-            util.callable(func)
+            callable(func)
             and func.__name__ == func_name
             and not func.__doc__
             and hasattr(list, func_name)
@@ -1326,7 +1355,7 @@ class _AssociationDict(_AssociationCollection):
         elif len(a) == 1:
             seq_or_map = a[0]
             # discern dict from sequence - took the advice from
-            # http://www.voidspace.org.uk/python/articles/duck_typing.shtml
+            # https://www.voidspace.org.uk/python/articles/duck_typing.shtml
             # still not perfect :(
             if hasattr(seq_or_map, "keys"):
                 for item in seq_or_map:
@@ -1370,7 +1399,7 @@ class _AssociationDict(_AssociationCollection):
 
     for func_name, func in list(locals().items()):
         if (
-            util.callable(func)
+            callable(func)
             and func.__name__ == func_name
             and not func.__doc__
             and hasattr(dict, func_name)
@@ -1589,7 +1618,7 @@ class _AssociationSet(_AssociationCollection):
 
     for func_name, func in list(locals().items()):
         if (
-            util.callable(func)
+            callable(func)
             and func.__name__ == func_name
             and not func.__doc__
             and hasattr(set, func_name)
